@@ -20,18 +20,61 @@ import httplib, urllib
 import libcloud
 from libcloud.types import NodeState, DeploymentError
 from libcloud.ssh import SSHClient
+from libcloud.httplib_ssl import LibcloudHTTPSConnection
+from httplib import HTTPConnection as LibcloudHTTPConnection
 import time
 import hashlib
 import StringIO
+import ssl
 import os
 import socket
 import struct
 from pipes import quote as pquote
 
-
 class Node(object):
     """
-    A Base Node class to derive from.
+    Provide a common interface for handling nodes of all types.
+
+    The Node object provides the interface in libcloud through which
+    we can manipulate nodes in different cloud providers in the same
+    way.  Node objects don't actually do much directly themselves,
+    instead the node driver handles the connection to the node.
+
+    You don't normally create a node object yourself; instead you use
+    a driver and then have that create the node for you.
+
+    >>> from libcloud.drivers.dummy import DummyNodeDriver
+    >>> driver = DummyNodeDriver(0)
+    >>> node = driver.create_node()
+    >>> node.public_ip[0]
+    '127.0.0.3'
+    >>> node.name
+    'dummy-3'
+
+    You can also get nodes from the driver's list_node function.
+
+    >>> node = driver.list_nodes()[0]
+    >>> node.name
+    'dummy-1'
+
+    the node keeps a reference to its own driver which means that we
+    can work on nodes from different providers without having to know
+    which is which.
+
+    >>> driver = DummyNodeDriver(72)
+    >>> node2 = driver.create_node()
+    >>> node.driver.creds
+    0
+    >>> node2.driver.creds
+    72
+
+    Althrough Node objects can be subclassed, this isn't normally
+    done.  Instead, any driver specific information is stored in the
+    "extra" proproperty of the node.
+
+    >>> node.extra
+    {'foo': 'bar'}
+
     """
 
     def __init__(self, id, name, state, public_ip, private_ip,
@@ -50,19 +93,67 @@ class Node(object):
 
     def get_uuid(self):
         """Unique hash for this node
+
         @return: C{string}
+
+        The hash is a function of an SHA1 hash of the node's ID and
+        its driver which means that it should be unique between all
+        nodes.  In some subclasses (e.g. GoGrid) there is no ID
+        available so the public IP address is used.  This means that,
+        unlike a properly done system UUID, the same UUID may mean a
+        different system install at a different time
+
+        >>> from libcloud.drivers.dummy import DummyNodeDriver
+        >>> driver = DummyNodeDriver(0)
+        >>> node = driver.create_node()
+        >>> node.get_uuid()
+        'd3748461511d8b9b0e0bfa0d4d3383a619a2bb9f'
+
+        Note, for example, that this example will always produce the
+        same UUID!
         """
         return hashlib.sha1("%s:%d" % (self.id,self.driver.type)).hexdigest()
 
     def reboot(self):
         """Reboot this node
+
         @return: C{bool}
+
+        This calls the node's driver and reboots the node
+
+        >>> from libcloud.drivers.dummy import DummyNodeDriver
+        >>> driver = DummyNodeDriver(0)
+        >>> node = driver.create_node()
+        >>> from libcloud.types import NodeState
+        >>> node.state == NodeState.RUNNING
+        True
+        >>> node.state == NodeState.REBOOTING
+        False
+        >>> node.reboot()
+        True
+        >>> node.state == NodeState.REBOOTING
+        True
         """
         return self.driver.reboot_node(self)
 
     def destroy(self):
         """Destroy this node
+
         @return: C{bool}
+
+        This calls the node's driver and destroys the node
+
+        >>> from libcloud.drivers.dummy import DummyNodeDriver
+        >>> driver = DummyNodeDriver(0)
+        >>> from libcloud.types import NodeState
+        >>> node = driver.create_node()
+        >>> node.state == NodeState.RUNNING
+        True
+        >>> node.destroy()
+        True
+        >>> node.state == NodeState.RUNNING
+        False
+
         """
         return self.driver.destroy_node(self)
 
@@ -76,6 +167,25 @@ class Node(object):
 class NodeSize(object):
     """
     A Base NodeSize class to derive from.
+
+    NodeSizes are objects which are typically returned a driver's
+    list_sizes function.  They contain a number of different
+    parameters which define how big an image is.
+
+    The exact parameters available depends on the provider.
+
+    N.B. Where a parameter is "unlimited" (for example bandwidth in
+    Amazon) this will be given as 0.
+
+    >>> from libcloud.drivers.dummy import DummyNodeDriver
+    >>> driver = DummyNodeDriver(0)
+    >>> size = driver.list_sizes()[0]
+    >>> size.ram
+    128
+    >>> size.bandwidth
+    500
+    >>> size.price
+    4
     """
 
     def __init__(self, id, name, ram, disk, bandwidth, price, driver):
@@ -95,7 +205,25 @@ class NodeSize(object):
 
 class NodeImage(object):
     """
-    A Base NodeImage class to derive from.
+    An operating system image.
+
+    NodeImage objects are typically returned by the driver for the
+    cloud provider in response to the list_images function
+
+    >>> from libcloud.drivers.dummy import DummyNodeDriver
+    >>> driver = DummyNodeDriver(0)
+    >>> image = driver.list_images()[0]
+    >>> image.name
+    'Ubuntu 9.10'
+
+    Apart from name and id, there is no further standard information;
+    other parameters are stored in a driver specific "extra" variable
+
+    When creating a node, a node image should be given as an argument
+    to the create_node function to decide which OS image to use.
+
+    >>> node = driver.create_node(image=image)
+
     """
 
     def __init__(self, id, name, driver, extra=None):
@@ -112,7 +240,13 @@ class NodeImage(object):
 
 class NodeLocation(object):
     """
-    A base NodeLocation class to derive from.
+    A physical location where nodes can be.
+
+    >>> from libcloud.drivers.dummy import DummyNodeDriver
+    >>> driver = DummyNodeDriver(0)
+    >>> location = driver.list_locations()[0]
+    >>> location.country
+    'US'
     """
 
     def __init__(self, id, name, country, driver):
@@ -127,6 +261,16 @@ class NodeLocation(object):
 class NodeAuthSSHKey(object):
     """
     An SSH key to be installed for authentication to a node.
+
+    This is the actual contents of the users ssh public key which will
+    normally be installed as root's public key on the node.
+
+    >>> pubkey = '...' # read from file
+    >>> from libcloud.base import NodeAuthSSHKey
+    >>> k = NodeAuthSSHKey(pubkey)
+    >>> k
+    <NodeAuthSSHKey>
+
     """
     def __init__(self, pubkey):
         self.pubkey = pubkey
@@ -257,13 +401,13 @@ class LoggingConnection():
         cmd.extend([pquote("https://%s:%d%s" % (self.host, self.port, url))])
         return " ".join(cmd)
 
-class LoggingHTTPSConnection(LoggingConnection, httplib.HTTPSConnection):
+class LoggingHTTPSConnection(LoggingConnection, LibcloudHTTPSConnection):
     """
     Utility Class for logging HTTPS connections
     """
 
     def getresponse(self):
-        r = httplib.HTTPSConnection.getresponse(self)
+        r = LibcloudHTTPSConnection.getresponse(self)
         if self.log is not None:
             r, rv = self._log_response(r)
             self.log.write(rv + "\n")
@@ -277,16 +421,15 @@ class LoggingHTTPSConnection(LoggingConnection, httplib.HTTPSConnection):
             self.log.write(pre +
                            self._log_curl(method, url, body, headers) + "\n")
             self.log.flush()
-        return httplib.HTTPSConnection.request(self, method, url,
-                                               body, headers)
+        return LibcloudHTTPSConnection.request(self, method, url, body, headers)
 
-class LoggingHTTPConnection(LoggingConnection, httplib.HTTPConnection):
+class LoggingHTTPConnection(LoggingConnection, LibcloudHTTPConnection):
     """
     Utility Class for logging HTTP connections
     """
 
     def getresponse(self):
-        r = httplib.HTTPConnection.getresponse(self)
+        r = LibcloudHTTPConnection.getresponse(self)
         if self.log is not None:
             r, rv = self._log_response(r)
             self.log.write(rv + "\n")
@@ -300,23 +443,15 @@ class LoggingHTTPConnection(LoggingConnection, httplib.HTTPConnection):
             self.log.write(pre +
                            self._log_curl(method, url, body, headers) + "\n")
             self.log.flush()
-        return httplib.HTTPConnection.request(self, method, url,
+        return LibcloudHTTPConnection.request(self, method, url,
                                                body, headers)
 
 class ConnectionKey(object):
     """
     A Base Connection class to derive from.
     """
-
-    # WARNING: Python's built-in SSL does not do certificate validation.  As
-    # such, one cannot be sure of the other end of the conversation with any
-    # sufficient authority.  If you are in a position to be exploited (i.e., on
-    # an untrusted network), be cautious with SSL connections.  This is an issue
-    # with upstream Python (see http://bugs.python.org/issue1589 for details)
-    # and not with libcloud.
-
-    #conn_classes = (httplib.LoggingHTTPConnection, LoggingHTTPSConnection)
-    conn_classes = (httplib.HTTPConnection, httplib.HTTPSConnection)
+    #conn_classes = (LoggingHTTPSConnection)
+    conn_classes = (LibcloudHTTPConnection, LibcloudHTTPSConnection)
 
     responseCls = Response
     connection = None
@@ -355,7 +490,9 @@ class ConnectionKey(object):
         host = host or self.host
         port = port or self.port[self.secure]
 
-        connection = self.conn_classes[self.secure](host, port)
+        kwargs = {'host': host, 'port': port}
+
+        connection = self.conn_classes[self.secure](**kwargs)
         # You can uncoment this line, if you setup a reverse proxy server
         # which proxies to your endpoint, and lets you easily capture
         # connections in cleartext when you setup the proxy to do SSL
@@ -439,8 +576,12 @@ class ConnectionKey(object):
         # Removed terrible hack...this a less-bad hack that doesn't execute a
         # request twice, but it's still a hack.
         self.connect()
-        self.connection.request(method=method, url=url, body=data,
-                                headers=headers)
+        try:
+            self.connection.request(method=method, url=url, body=data,
+                                    headers=headers)
+        except ssl.SSLError, e:
+            raise ssl.SSLError(str(e))
+
         response = self.responseCls(self.connection.getresponse())
         response.connection = self
         return response
@@ -487,6 +628,11 @@ class ConnectionUserAndKey(ConnectionKey):
 class NodeDriver(object):
     """
     A base NodeDriver class to derive from
+
+    This class is always subclassed by a specific driver.  For
+    examples of base behavior of most functions (except deploy node)
+    see the dummy driver.
+
     """
 
     connectionCls = ConnectionKey
@@ -630,15 +776,40 @@ class NodeDriver(object):
         Depends on a Provider Driver supporting either using a specific password
         or returning a generated password.
 
-        This function may raise a L{DeplyomentException}, if a create_node
-        call was successful, but there is a later error (like SSH failing or 
+        This function may raise a L{DeploymentException}, if a create_node
+        call was successful, but there is a later error (like SSH failing or
         timing out).  This exception includes a Node object which you may want
         to destroy if incomplete deployments are not desirable.
 
         @keyword    deploy: Deployment to run once machine is online and availble to SSH.
         @type       deploy: L{Deployment}
 
+        @keyword    ssh_username: Optional name of the account which is used when connecting to
+                                  SSH server (default is root)
+        @type       ssh_username: C{str}
+
+        @keyword    ssh_port: Optional SSH server port (default is 22)
+        @type       ssh_port: C{int}
+
         See L{NodeDriver.create_node} for more keyword args.
+
+        >>> from libcloud.drivers.dummy import DummyNodeDriver
+        >>> from libcloud.deployment import ScriptDeployment, MultiStepDeployment
+        >>> from libcloud.base import NodeAuthSSHKey
+        >>> driver = DummyNodeDriver(0)
+        >>> key = NodeAuthSSHKey('...') # read from file
+        >>> script = ScriptDeployment("yum -y install emacs strace tcpdump")
+        >>> msd = MultiStepDeployment([key, script])
+        >>> def d():
+        ...     try:
+        ...         node = driver.deploy_node(deploy=msd)
+        ...     except NotImplementedError:
+        ...         print "not implemented for dummy driver"
+        >>> d()
+        not implemented for dummy driver
+
+        Deploy node is typically not overridden in subclasses.  The
+        existing implementation should be able to handle most such.
         """
         # TODO: support ssh keys
         WAIT_PERIOD=3
@@ -675,9 +846,12 @@ class NodeDriver(object):
               if node.public_ip is not None and node.public_ip != "" and node.state == NodeState.RUNNING:
                   break
 
+          ssh_username = kwargs.get('ssh_username', 'root')
+          ssh_port = kwargs.get('ssh_port', 22)
+
           client = SSHClient(hostname=node.public_ip[0],
-                              port=22, username='root',
-                              password=password)
+                             port=ssh_port, username=ssh_username,
+                             password=password)
           laste = None
           while time.time() < end:
               laste = None
@@ -721,3 +895,8 @@ def is_private_subnet(ip):
             return True
 
     return False
+
+
+if __name__ == "__main__":
+    import doctest
+    doctest.testmod()
