@@ -18,10 +18,12 @@ Drivers for working with different providers
 """
 
 __all__ = [
+    'brightbox',
     'dummy',
     'ec2',
     'ecp',
     'elastichosts',
+    'cloudsigma',
     'gogrid',
     'ibm_sbc',
     'linode',
